/**
 * @file fleet.c
 *
 * @brief Handles the fleet stuff.
 */

#include "fleet.h"

#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <limits.h>

#include "lxml.h"
#include "lephisto.h"
#include "log.h"
#include "pilot.h"
#include "ldata.h"

#define FLEET_DATA      "../dat/fleet.xml" /**< Where to find fleet data. */
#define FLEETGROUP_DATA "../dat/fleetgroup.xml" /**< Where to find fleetgroup. */

#define CHUNK_SIZE  32  /**< Size to allocate memory by. */

/* Stack of fleets. */
static Fleet* fleet_stack = NULL; /**< Fleet stack. */
static int nfleets = 0; /**< Number of fleets. */

/* Stack of fleetgroups. */
static FleetGroup* fleetgroup_stack = NULL; /**< FleetGroup stack. */
static int nfleetgroups = 0; /**< Number of fleetgroups. */

static int fleet_parse(Fleet* tmp, const xmlNodePtr parent);
static int fleet_parseGroup(FleetGroup* fltgrp, xmlNodePtr parent);

/**
 * @brief Grabs a fleet out of the stack.
 *    @param name Name of the fleet to match.
 *    @return The fleet mactching name or NULL if not found.
 */
Fleet* fleet_get(const char* name) {
  int i;

  for(i = 0; i < nfleets; i++)
    if(strcmp(fleet_stack[i].name, name)==0)
      return &fleet_stack[i];

  return NULL;
}

/**
 * @brief Grab a fleetgroup out of the stack.
 *    @param name Name of the fleetgroup to match.
 *    @return The fleetgroup matching name or NULL if not found.
 */
FleetGroup* fleet_getGroup(const char* name) {
  int i;

  for(i = 0; i < nfleetgroups; i++)
    if(strcmp(fleetgroup_stack[i].name, name)==0)
      return &fleetgroup_stack[i];

  return NULL;
}

/**
 * @brief Parses the fleet node.
 *    @param tmp Fleet to load.
 *    @param parent Parent xml node of the fleet in question.
 *    @return A newly allocated fleet loaded with data in parent node.
 */
static int fleet_parse(Fleet* tmp, const xmlNodePtr parent) {
  xmlNodePtr cur, node;
  FleetPilot* pilot;
  char* c;
  int mem;
  node = parent->xmlChildrenNode;

  /* Sane defaults and clean up. */
  memset(tmp, 0, sizeof(Fleet));
  tmp->faction = -1;

  tmp->name = (char*)xmlGetProp(parent, (xmlChar*)"name"); /* Already mallocs. */
  if(tmp->name == NULL) WARN("Fleet in "FLEET_DATA" has invalid or no name");

  do { /* Load all the data. */
    if(xml_isNode(node, "faction"))
      tmp->faction = faction_get(xml_get(node));
    else if(xml_isNode(node, "ai"))
      tmp->ai = xml_getStrd(node);
    else if(xml_isNode(node, "pilots")) {
      cur = node->children;
      mem = 0;
      do {
        if(xml_isNode(cur, "pilot")) {
          /* See if must grow. */
          tmp->npilots++;
          if(tmp->npilots > mem) {
            mem += CHUNK_SIZE;
            tmp->pilots = realloc(tmp->pilots, mem * sizeof(FleetPilot));
          }
          pilot = &tmp->pilots[tmp->npilots-1];

          /* Clear memory. */
          memset(pilot, 0, sizeof(FleetPilot));

          /* Check for name override. */
          xmlr_attr(cur, "name", c);
          pilot->name = c; /* No need to free since it will have later. */

          /* Check for ai override. */
          xmlr_attr(cur, "ai", pilot->ai);

          /* Load pilots ship. */
          pilot->ship = ship_get(xml_get(cur));
          if(pilot->ship == NULL)
            WARN("Pilot %s in Fleet %s has null ship", pilot->name, tmp->name);

          /* Load chance. */
          xmlr_attr(cur, "chance", c);
          pilot->chance = atoi(c);
          if(pilot->chance == 0)
            WARN("Pilot %s in Fleet %s has 0%% chance of appearing",
                pilot->name, tmp->name);

          if(c != NULL)
            free(c); /* Free the external malloc. */
        }
      } while(xml_nextNode(cur));

      /* Resize to minimum. */
      tmp->pilots = realloc(tmp->pilots, sizeof(FleetPilot)*tmp->npilots);
    }
  } while(xml_nextNode(node));

#define MELEMENT(o,s) \
  if(o) WARN("Fleet '%s' missing '"s"' element", tmp->name)
  /**< Hack to check for missing fields. */
  MELEMENT(tmp->ai        == NULL,  "ai");
  MELEMENT(tmp->faction   == -1,    "faction");
  MELEMENT(tmp->pilots    == NULL,  "pilots");
#undef MELEMENT

  return 0;
}

/**
 * @brief Loads all the fleets.
 *    @return 0 on success.
 */
static int fleet_loadFleets(void) {
  int mem;
  uint32_t bufsize;

  char* buf;
  xmlNodePtr node;

  xmlDocPtr doc;

  /* Load the data. */
  buf = ldata_read(FLEET_DATA, &bufsize);
  doc = xmlParseMemory(buf, bufsize);

  node = doc->xmlChildrenNode; /* Fleets node. */
  if(strcmp((char*)node->name, "Fleets")) {
    ERR("Malformed "FLEET_DATA" file: missing root element 'Fleets'.");
    return -1;
  }

  node = node->xmlChildrenNode; /* First fleet node. */
  if(node == NULL) {
    ERR("Malformed "FLEET_DATA" file: does not contain elements.");
    return -1;
  }

  mem = 0;
  do {
    if(xml_isNode(node, "fleet")) {
      /* See if memory must grow. */
      nfleets++;
      if(nfleets > mem) {
        mem += CHUNK_SIZE;
        fleet_stack = realloc(fleet_stack, sizeof(Fleet) * mem);
      }

      /* Load the fleet. */
      fleet_parse(&fleet_stack[nfleets-1], node);
    }
  } while(xml_nextNode(node));

  /* Shrink to minimum. */
  fleet_stack = realloc(fleet_stack, sizeof(Fleet) * nfleets);

  xmlFreeDoc(doc);
  free(buf);

  return 0;
}

/**
 * @brief Parses a fleetgroup from an xml node.
 *    @param fltgrp FleetGroup to fill with data from the xml node.
 *    @param parent Node containing fleetgroup data.
 *    @return 0 on success.
 */
static int fleet_parseGroup(FleetGroup* fltgrp, xmlNodePtr parent) {
  int mem;
  xmlNodePtr node;
  Fleet* f;
  char* buf;

  /* Clear memory. */
  memset(fltgrp, 0, sizeof(FleetGroup));

  /* Get the name. */
  xmlr_attr(parent, "name", fltgrp->name);

  /* Load the fleetgroup data. */
  node = parent->children;
  mem = 0;
  do {
    if(xml_isNode(node, "fleet")) {
      /* See if memory must grow. */
      fltgrp->nfleets++;
      if(fltgrp->nfleets > mem) {
        mem += CHUNK_SIZE;
        fltgrp->fleets = realloc(fltgrp->fleets, sizeof(Fleet*) * mem);
        fltgrp->chance = realloc(fltgrp->chance, sizeof(int) * mem);
      }

      /* Add the fleet. */
      f = fleet_get(xml_get(node));
      if(f == NULL) {
        WARN("Fleet '%s' in FleetGroup '%s' not found in stack.",
            xml_get(node), fltgrp->name);
        fltgrp->nfleets--;
        continue;
      }
      fltgrp->fleets[fltgrp->nfleets-1] = f;

      /* Get the chance. */
      xmlr_attr(node, "chance", buf);
      if(buf == NULL) {
        WARN("Fleet '%s' in FleetGroup '%s' missing 'chance' attribute.",
            xml_get(node), fltgrp->name);
        fltgrp->chance[fltgrp->nfleets-1] = 0;
        continue;
      }
      fltgrp->chance[fltgrp->nfleets-1] = CLAMP(0, 100, atoi(buf));
    }
  } while(xml_nextNode(node));

  return 0;
}

/**
 * @brief Load all the fleetgroups.
 *    @return 0 on success.
 */
static int fleet_loadFleetGroups(void) {
  int mem;
  uint32_t bufsize;
  char* buf;
  xmlNodePtr node;
  xmlDocPtr doc;

  /* Create the document. */
  buf = ldata_read(FLEETGROUP_DATA, &bufsize);
  doc = xmlParseMemory(buf, bufsize);

  node = doc->xmlChildrenNode; /* Fleetgroups node. */
  if(strcmp((char*)node->name, "FleetGroups")) {
    ERR("Malformed "FLEETGROUP_DATA" file: missing root element 'FleetGroups'");
    return -1;
  }

  node = node->xmlChildrenNode; /* First fleetgroup node. */
  if(node == NULL) {
    ERR("Malformed "FLEETGROUP_DATA" file: does not contain elements");
    return -1;
  }

  mem = 0;
  do {
    if(xml_isNode(node, "fleetgroup")) {
      /* See if memory must grow. */
      nfleetgroups++;
      if(nfleets > mem) {
        mem += CHUNK_SIZE;
        fleetgroup_stack = realloc(fleetgroup_stack, sizeof(FleetGroup) * mem);
      }

      /* Load the fleetgroup. */
      fleet_parseGroup(&fleetgroup_stack[nfleetgroups-1], node);
    }
  } while(xml_nextNode(node));

  /* Shrink to minimum. */
  fleet_stack = realloc(fleet_stack, sizeof(Fleet) * nfleets);

  xmlFreeDoc(doc);
  free(buf);

  return 0;
}

/**
 * @brief Load all the fleets and fleetgroups.
 *    @return 0 on success.
 */
int fleet_load(void) {
  if(fleet_loadFleets())
    return -1;
  if(fleet_loadFleetGroups())
    return -1;

  DEBUG("Loaded %d Fleet%s", nfleets, (nfleets==1) ? "" : "s");

  return 0;
}

/**
 * @brief Cleans up by freeing all the fleet data.
 */
void fleet_free(void) {
  int i,j;
  if(fleet_stack != NULL) {
    for(i = 0; i < nfleets; i++) {
      for(j = 0; j < fleet_stack[i].npilots; j++) {
        if(fleet_stack[i].pilots[j].name)
          free(fleet_stack[i].pilots[j].name);
        if(fleet_stack[i].pilots[j].ai)
          free(fleet_stack[i].pilots[j].ai);
      }
      free(fleet_stack[i].name);
      free(fleet_stack[i].pilots);
      free(fleet_stack[i].ai);
    }
    free(fleet_stack);
  }
  fleet_stack = NULL;
  nfleets = 0;

  /* Free the fleetgroup stack. */
  if(fleetgroup_stack != NULL) {
    for(i = 0; i < nfleetgroups; i++) {
      free(fleetgroup_stack[i].name);
      free(fleetgroup_stack[i].fleets);
      free(fleetgroup_stack[i].chance);
    }
    free(fleetgroup_stack);
  }
  fleetgroup_stack = NULL;
  nfleetgroups = 0;
}

